/* bbowda - Black Box Optimization With Data Analysis
   Copyright (C) 2006-2007 Kevin Kofler <Kevin@tigcc.ticalc.org>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version. A copy of the GNU General Public
   License version 3 can be found in the file gpl-3.0.txt.

   Linking bbowda statically or dynamically (directly or indirectly) with
   other modules is making a combined work based on bbowda. Thus, the terms
   and conditions of the GNU General Public License cover the whole
   combination.

   In addition, as a special exception, the copyright holder of bbowda gives
   you permission to combine the bbowda program:
   * with free software programs or libraries that are released under the
     GNU Library or Lesser General Public License (LGPL), either version 2
     of the License, or (at your option) any later version,
   * with free software programs or libraries that are released under the
     IBM Common Public License (CPL), either version 1.0 of the License, or
     (at your option) any later version,
   * with code included in the standard release of MUMPS under the MUMPS
     Conditions of Use as reproduced in licenses.txt (or modified versions
     of such code, with unchanged license) and
   * if you qualify for a free of charge license of DONLP2, with code
     included in the standard release of DONLP2 under the DONLP2 Conditions
     of Use as reproduced in licenses.txt (or modified versions of such code,
     with unchanged license).

   You may copy and distribute such a system following the terms of the GNU
   GPL for bbowda and the licenses of the other code concerned, provided that
   you include the source code of that other code when and as the GNU GPL
   requires distribution of source code.

   Note that people who make modified versions of bbowda are not obligated
   to grant this special exception for their modified versions; it is their
   choice whether to do so. The GNU General Public License gives permission
   to release a modified version without this exception; this exception also
   makes it possible to release a modified version which carries forward
   this exception.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define _ISOC99_SOURCE
#include "covar.h"
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "eval.h"
#include "nlpopt.h"
#include "xmalloc.h"

#ifdef __GNUC__
#define ATTR_UNUSED __attribute__((unused))
#else
#define ATTR_UNUSED /**/
#endif

/* Cholesky factor of the covariance matrix */
static DOUBLE L[DIMX+DIMZ+DIMY+DIMY_EQ][DIMX+DIMZ+DIMY+DIMY_EQ];
/* Transpose of L, reversed, to avoid cache misses in compute_MX */
static DOUBLE LTrev[DIMX+DIMZ+DIMY+DIMY_EQ][DIMX+DIMZ+DIMY+DIMY_EQ];
/* Copy of currpts sorted by the norm of the equality constraint violation */
static DOUBLE (*usedpts)[DIMX+DIMY+DIMY_EQ]=NULL;
/* Full number of points in usedpts */
static size_t usedpts_size;
/* Number of points in usedpts which should actually be used */
static size_t numusedpts;

/* Global variables */
DOUBLE Xbar[DIMX+DIMZ+DIMY+DIMY_EQ];
DOUBLE klow;
DOUBLE kup;

static void for_each_used(DOUBLE *acc, DOUBLE *aux, void (*update)(const DOUBLE *, DOUBLE *, DOUBLE *))
{
  size_t j;
  for (j=0; j<numusedpts; j++) {
    update(usedpts[j],acc,aux);
  }
}

/* MX = C^-1 X = (L LT)^-1 X = L^-T L^-1 X */
void compute_MX(const DOUBLE *X, DOUBLE *MX)
{
  size_t i,j,k;
  /* MX := L^-1 X */
  for (i=0; i<(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ); i++) {
    MX[i]=X[i];
    for (j=0; j<i; j++) {
      MX[i]-=L[i][j]*MX[j];
    }
    MX[i]/=L[i][i];
  }
  /* MX := L^-T X */
  /* try hard to avoid cache misses */
  for (k=0; k<(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ); k++) {
    j=i;
    DOUBLE d=LTrev[k][--i];
    for (; j<(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ); j++) {
      MX[i]-=LTrev[k][j]*MX[j];
    }
    MX[i]/=d;
  }
}

/* empirical formula for the weight:
   * our point weights decrease with the 6th power of the distance
   * the worse the point, the less it will be weighted
   weight = (||x-xbest||^2)^(-6)/(1/10+(goodness-best_goodness))^(1/2) */
static DOUBLE weight_of_point(const DOUBLE *x)
{
  size_t i;
  DOUBLE goodness=get_point_goodness(x), weight=0.;
  for (i=0; i<(size_t)DIMX; i++) {
    DOUBLE d=best_point[i]-x[i];
    weight += d*d;
  }
  if (weight == 0.) return 0.;
  weight = pow(weight,-3.);
  weight /= sqrt(.1+(goodness-best_goodness));
  return weight;
}

static void compute_covar(const DOUBLE *x, DOUBLE *Xbar, DOUBLE *count)
{
  size_t i,j,k,l,a,b;
  DOUBLE weight=weight_of_point(x);
  for (i=0; i<(size_t)DIMX; i++) {
    for (k=0; k<=i; k++) {
      L[i][k] += weight*(x[i]-Xbar[i])*(x[k]-Xbar[k]);
    }
  }
  for (i=0,a=(size_t)DIMX; i<(size_t)DIMX; i++) {
    for (j=0; j<=i; j++,a++) {
      for (k=0; k<(size_t)DIMX; k++) {
        L[a][k] += weight*(x[i]*x[j]-Xbar[a])*(x[k]-Xbar[k]);
      }
      for (k=0,b=(size_t)DIMX; b<=a; k++) {
        for (l=0; l<=k && b<=a; l++,b++) {
          L[a][b] += weight*(x[i]*x[j]-Xbar[a])*(x[k]*x[l]-Xbar[b]);
        }
      }
    }
  }
  for (i=0; i<(size_t)(DIMY+DIMY_EQ); i++) {
    for (k=0; k<(size_t)DIMX; k++) {
      L[i+DIMX+DIMZ][k] += weight*(x[i+DIMX]-Xbar[i+DIMX+DIMZ])*(x[k]-Xbar[k]);
    }
    for (k=0,b=(size_t)DIMX; k<(size_t)DIMX; k++) {
      for (l=0; l<=k; l++,b++) {
        L[i+DIMX+DIMZ][b] += weight*(x[i+DIMX]-Xbar[i+DIMX+DIMZ])*(x[k]*x[l]-Xbar[b]);
      }
    }
    for (k=0; k<=i; k++) {
      L[i+DIMX+DIMZ][k+DIMX+DIMZ] += weight*(x[i+DIMX]-Xbar[i+DIMX+DIMZ])*(x[k+DIMX]-Xbar[k+DIMX+DIMZ]);
    }
  }
  (*count)+=weight;
}

static void compute_kup_klow(const DOUBLE *x, DOUBLE *dmax, DOUBLE *aux ATTR_UNUSED)
{
  /* compute k = XT M X */
  DOUBLE X[DIMX+DIMZ+DIMY+DIMY_EQ], MX[DIMX+DIMZ+DIMY+DIMY_EQ], *p=X, k=0., d=0.;
  size_t i,j;
  for (i=0; i<(size_t)DIMX; i++) {
    d += (x[i]-best_point[i])*(x[i]-best_point[i]);
  }
  if (d>*dmax) return;
  for (i=0; i<(size_t)DIMX; i++) {
    *(p++) = x[i];
  }
  for (i=0; i<(size_t)DIMX; i++) {
    for (j=0; j<=i; j++) {
      *(p++) = x[i]*x[j];
    }
  }
  for (i=0; i<(size_t)(DIMY+DIMY_EQ); i++) {
    *(p++) = x[DIMX+i];
  }
  for (i=0; i<(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ); i++) {
    X[i] -= Xbar[i];
  }
  compute_MX(X,MX);
  for (i=0; i<(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ); i++) {
    k += X[i]*MX[i];
  }

  /* update kup, klow */
  if (k>kup) kup=k;
  if (k<klow) klow=k;
}

static int compare_DOUBLE(const void *p, const void *q)
{
  return (*(DOUBLE *)p>*(DOUBLE *)q)-(*(DOUBLE *)p<*(DOUBLE *)q);
}

static int compare_yeq(const void *p, const void *q)
{
  size_t i;
  DOUBLE yeq1=0., yeq2=0.;
  // compute norm(p[yeq])^2 and norm(q[yeq])^2
  for (i=(size_t)(DIMX+DIMY); i<(size_t)(DIMX+DIMY+DIMY_EQ); i++) {
    yeq1+=i[(DOUBLE*)p]*i[(DOUBLE*)p];
    yeq2+=i[(DOUBLE*)q]*i[(DOUBLE*)q];
  }
  // compare them
  return (yeq1>yeq2)-(yeq1<yeq2);
}

void build_local_regcovar_model(void)
{
  if (DIMY_EQ) {
    if (usedpts) {
      /* make room for the new points in usedpts */
      usedpts=xrealloc(usedpts,numcurrpts*((size_t)(DIMX+DIMY+DIMY_EQ))*sizeof(DOUBLE));
      /* sort the new points by the norm of the equality constraint violation
         using insertion sort */
      size_t i,j;
      for (i=usedpts_size; i<numcurrpts; i++) {
        /* This could be done more efficiently using binary search, but the
           memmove is O(n) anyway. We can't use bsearch as it only returns exact
           matches. */
        for (j=0; j<i; j++) {
          if (compare_yeq(currpts[i],usedpts[j])>=0) break;
        }
        memmove(usedpts+j+1,usedpts+j,(i-j)*((size_t)(DIMX+DIMY+DIMY_EQ))*sizeof(DOUBLE));
        memcpy(usedpts[j],currpts[i],((size_t)(DIMX+DIMY+DIMY_EQ))*sizeof(DOUBLE));
      }
    } else {
      /* allocate a copy of currpts */
      usedpts=xmalloc(numcurrpts*((size_t)(DIMX+DIMY+DIMY_EQ))*sizeof(DOUBLE));
      memcpy(usedpts,currpts,numcurrpts*((size_t)(DIMX+DIMY+DIMY_EQ))*sizeof(DOUBLE));
      /* sort the points by the norm of the equality constraint violation */
      qsort(usedpts,numcurrpts,(size_t)(DIMX+DIMY+DIMY_EQ)*sizeof(DOUBLE),compare_yeq);
    }
    usedpts_size=numcurrpts;
    /* use only the first half unless we don't have enough points */
    numusedpts=(numcurrpts>=28)?(numcurrpts>>1):numcurrpts;
  } else {
    numusedpts=numcurrpts;
    usedpts=currpts;
  }

  /* compute the covariance matrix */
  size_t i,j,k;
  DOUBLE n=0., Cdiag[DIMX+DIMZ+DIMY+DIMY_EQ], d[numusedpts], dmax;
  for (k=0; k<(size_t)DIMX; k++) {
    Xbar[k]=best_point[k];
  }
  for (i=0; i<(size_t)DIMX; i++) {
    for (j=0; j<=i; j++) {
      Xbar[k++]=best_point[i]*best_point[j];
    }
  }
  for (k=0; k<(size_t)(DIMY+DIMY_EQ); k++) {
    Xbar[k+DIMX+DIMZ]=best_point[k+DIMX];
  }
  for (i=0; i<(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ); i++) {
    for (j=0; j<=i; j++) {
      L[i][j]=0.;
    }
  }
  for_each_used(Xbar,&n,compute_covar);
  for (i=0; i<(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ); i++) {
    for (j=0; j<=i; j++) {
      L[i][j]/=n;
    }
  }

  /* save the diagonal of the matrix */
  for (j=0; j<(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ); j++) {
    Cdiag[j]=L[j][j];
  }
  /* compute the regularized Cholesky factorization of the matrix */
  for (j=0; j<(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ); j++) {
    DOUBLE Ljj;
    /* regularize */
    if (L[j][j]<=sqrt(DBL_EPSILON)*Cdiag[j]) {
      L[j][j]=(Cdiag[j]==0.)?1.:(sqrt(DBL_EPSILON)*Cdiag[j]);
    }
    /* in higher dimensions, regularize harder */
    if (DIMX > 6)
      L[j][j]+=sqrt(DBL_EPSILON)*Cdiag[j];

    Ljj=L[j][j];
    for (i=j+1; i<(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ); i++) {
      for (k=i; k<(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ); k++) {
        L[k][i]-=L[k][j]*L[i][j]/Ljj;
      }
    }
    for (k=j; k<(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ); k++) {
      L[k][j]/=sqrt(Ljj);
    }
  }
  /* compute the reversed transpose */
  for (i=0; i<(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ); i++) {
    for (j=0; j<=i; j++) {
      LTrev[(size_t)(DIMX+DIMZ+DIMY+DIMY_EQ-1)-j][i]=L[i][j];
    }
  }

  if (numusedpts<=(DIMX<<1)) {
    dmax=INFINITY;
  } else {
    /* compute distances from best point */
    for (j=0; j<numusedpts; j++) {
      d[j]=0.;
      for (i=0; i<(size_t)DIMX; i++) {
        d[j]+=(usedpts[j][i]-best_point[i])*(usedpts[j][i]-best_point[i]);
      }
    }
    /* sort them */
    qsort(d,numusedpts,sizeof(DOUBLE),compare_DOUBLE);
    /* take the 2*DIMX+1st entry as dmax, i.e. take at least 2*DIMX+1 points into
       account */
    dmax=d[DIMX<<1];
  }

  /* compute kup, klow */
  kup=-INFINITY;
  klow=INFINITY;
  for_each_used(&dmax,NULL,compute_kup_klow);
  /* fudge kup for higher dimensions so we get reasonably-sized trust regions,
     otherwise the local search gets stuck in a non-optimal point */
  kup *= pow(2.,(DOUBLE)(DIMX-1));

  /* set the NLP solver to optimize the local covariance model */
  optimization_problem=0;
}
