/* bbowda - Black Box Optimization With Data Analysis
   Copyright (C) 2006-2007 Kevin Kofler <Kevin@tigcc.ticalc.org>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version. A copy of the GNU General Public
   License version 3 can be found in the file gpl-3.0.txt.

   Linking bbowda statically or dynamically (directly or indirectly) with
   other modules is making a combined work based on bbowda. Thus, the terms
   and conditions of the GNU General Public License cover the whole
   combination.

   In addition, as a special exception, the copyright holder of bbowda gives
   you permission to combine the bbowda program:
   * with free software programs or libraries that are released under the
     GNU Library or Lesser General Public License (LGPL), either version 2
     of the License, or (at your option) any later version,
   * with free software programs or libraries that are released under the
     IBM Common Public License (CPL), either version 1.0 of the License, or
     (at your option) any later version,
   * with code included in the standard release of MUMPS under the MUMPS
     Conditions of Use as reproduced in licenses.txt (or modified versions
     of such code, with unchanged license) and
   * if you qualify for a free of charge license of DONLP2, with code
     included in the standard release of DONLP2 under the DONLP2 Conditions
     of Use as reproduced in licenses.txt (or modified versions of such code,
     with unchanged license).

   You may copy and distribute such a system following the terms of the GNU
   GPL for bbowda and the licenses of the other code concerned, provided that
   you include the source code of that other code when and as the GNU GPL
   requires distribution of source code.

   Note that people who make modified versions of bbowda are not obligated
   to grant this special exception for their modified versions; it is their
   choice whether to do so. The GNU General Public License gives permission
   to release a modified version without this exception; this exception also
   makes it possible to release a modified version which carries forward
   this exception.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define _ISOC99_SOURCE
#include "eval.h"
#include "problem.h"
#include "eqconst.h"
#include "xmalloc.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

size_t numcurrpts;
DOUBLE (*currpts)[DIMX+DIMY+DIMY_EQ];
DOUBLE best_goodness;
DOUBLE *best_point;

void init_points(void)
{
  size_t i;
  numcurrpts=NUMINITPTS;
  currpts=xmalloc((size_t)NUMINITPTS*((size_t)(DIMX+DIMY+DIMY_EQ))*sizeof(DOUBLE));
  for (i=0; i<(size_t)NUMINITPTS; i++) {
    memcpy(currpts[i],initpts[i],DIMX*sizeof(DOUBLE));
    evaluate_F(initpts[i],currpts[i]+DIMX);
  }
}

void new_point(const DOUBLE *x)
{
  size_t i=numcurrpts++;
  currpts=xrealloc(currpts,numcurrpts*((size_t)(DIMX+DIMY+DIMY_EQ))*sizeof(DOUBLE));
  memcpy(currpts[i],x,DIMX*sizeof(DOUBLE));
  evaluate_F(x,currpts[i]+DIMX);
  /* compute global over-/underestimates of equality constraints around the new point */
  compute_global_eq_cst_estimates_around(currpts[i]);
}

DOUBLE get_optimum_x(void)
{
  size_t j;
  DOUBLE optimum=INFINITY;
#ifdef OPTIMUM_TOL
  const DOUBLE tol=OPTIMUM_TOL;
#else
  const DOUBLE tol=sqrt(DBL_EPSILON);
#endif
  for (j=0; j<numcurrpts; j++) {
    DOUBLE cTx=0.;
    DOUBLE *x=currpts[j];
    size_t i;
    for (i=0; i<(size_t)DIMX; i++) {
      if (x[i]<xlow[i] || x[i]>xup[i]) {
        goto infeasible; /* infeasible point */
      }
    }
    for (i=0; i<(size_t)DIMY; i++) {
      if (x[DIMX+i]<Flow[i]-tol || x[DIMX+i]>Fup[i]+tol) {
        goto infeasible; /* infeasible point */
      }
    }
    for (i=0; i<(size_t)DIMY_EQ; i++) {
      if (fabs(x[DIMX+DIMY+i])>=tol) {
        goto infeasible; /* infeasible point */
      }
    }
    for (i=0; i<(size_t)(DIMX+DIMY); i++) {
      cTx+=c[i]*x[i];
    }
    if (cTx<optimum) {
      memcpy(optimum_x,x,(size_t)DIMX*sizeof(DOUBLE));
      optimum=cTx;
    }
    infeasible:;
  }
  return optimum;
}

/* The weights are determined using a penalty function approach, with a penalty
   term increasing over time:
   min cT x + numcurrpts * constraint_violation */
static DOUBLE get_point_penalty(const DOUBLE *x)
{
  DOUBLE penalty=0.;
  size_t i;
  for (i=0; i<(size_t)DIMX; i++) {
    if (x[i]<xlow[i]) {
      penalty+=numcurrpts*(xlow[i]-x[i]);
    } else if (x[i]>xup[i]) {
      penalty+=numcurrpts*(x[i]-xup[i]);
    }
  }
  for (i=0; i<(size_t)DIMY; i++) {
    if (x[DIMX+i]<Flow[i]) {
      penalty+=numcurrpts*(Flow[i]-x[DIMX+i]);
    } else if (x[DIMX+i]>Fup[i]) {
      penalty+=numcurrpts*(x[DIMX+i]-Fup[i]);
    }
  }
  for (i=0; i<(size_t)DIMY_EQ; i++) {
    penalty+=(numcurrpts<18?sqrt(numcurrpts/18.):
                  numcurrpts<77?numcurrpts/18.:
                  numcurrpts*sqrt(numcurrpts)/158.)*fabs(x[DIMX+DIMY+i]);
  }
  return penalty;
}

static DOUBLE get_point_constraint_violation(const DOUBLE *x)
{
  DOUBLE penalty=0.;
  size_t i;
  for (i=0; i<(size_t)DIMX; i++) {
    if (x[i]<xlow[i]) {
      penalty+=(xlow[i]-x[i]);
    } else if (x[i]>xup[i]) {
      penalty+=(x[i]-xup[i]);
    }
  }
  for (i=0; i<(size_t)DIMY; i++) {
    if (x[DIMX+i]<Flow[i]) {
      penalty+=(Flow[i]-x[DIMX+i]);
    } else if (x[DIMX+i]>Fup[i]) {
      penalty+=(x[DIMX+i]-Fup[i]);
    }
  }
  for (i=0; i<(size_t)DIMY_EQ; i++) {
    penalty+=fabs(x[DIMX+DIMY+i]);
  }
  return penalty;
}

static DOUBLE get_point_cTx(const DOUBLE *x)
{
  DOUBLE cTx=0.;
  size_t i;
  for (i=0; i<(size_t)(DIMX+DIMY); i++) {
    cTx+=c[i]*x[i];
  }
  return cTx;
}

DOUBLE get_point_goodness(const DOUBLE *x)
{
  return get_point_cTx(x)+get_point_penalty(x);
}

void get_best_point(void)
{
#ifdef USE_PENALTY_APPROACH
  size_t j;
  DOUBLE optimum=INFINITY;
  best_point=NULL;
  for (j=0; j<numcurrpts; j++) {
    DOUBLE cTx_penalty=get_point_goodness(currpts[j]);
    if (cTx_penalty<optimum) {
      best_point=currpts[j];
      optimum=cTx_penalty;
    }
  }
  if (!best_point) best_point=currpts[0];
  best_goodness=optimum;
#else /* filter method */
  static unsigned char *used=NULL;
  static size_t used_size=0;
  /* use the penalty to compute the optimum goodness, it's used for weighting */
  DOUBLE cTx[numcurrpts], penalty[numcurrpts];
  size_t i,j;
  DOUBLE optimum=INFINITY;
  best_point=NULL;
  /* new points are not used */
  used=xrealloc(used,numcurrpts);
  memset(used+used_size,0,numcurrpts-used_size);
  used_size=numcurrpts;
  for (j=0; j<numcurrpts; j++) {
    DOUBLE cTx_j=get_point_cTx(currpts[j]);
    DOUBLE penalty_j=get_point_penalty(currpts[j]);
    DOUBLE cTx_penalty=cTx_j+penalty_j;
    cTx[j]=cTx_j;
    penalty[j]=penalty_j;
    if (cTx_penalty<optimum) {
      optimum=cTx_penalty;
    }
  }
  best_goodness=optimum;
  /* now pick a "best point" using the filter approach */
  unsigned char pareto_optimal[numcurrpts];
  size_t num_pareto_points=0;
  memset(pareto_optimal,1,numcurrpts);
  for (i=0; i<numcurrpts; i++) {
    for (j=0; j<numcurrpts; j++) {
      /* if j is strictly better than i, i is not Pareto-optimal */
      if (j!=i && pareto_optimal[j] /* avoid unnecessary comparisons */ &&
          cTx[j] <= cTx[i] && penalty[j] <= penalty[i] &&
          (cTx[j] < cTx[i] || penalty[j] < penalty[i])) {
        pareto_optimal[i]=0;
        goto dont_count;
      }
    }
    num_pareto_points++;
    dont_count:;
  }
  /* pick a pseudorandom Pareto-optimal point as the best point
     pick a later one with a higher probability */
  /* find the most recent Pareto-optimal point */
  /* i is unsigned and will overflow to ULONG_MAX */
  for (i=numcurrpts-1; i<numcurrpts; i--) {
    if (pareto_optimal[i]) break;
  }
  /* if it is not used, pick it with a probability of 1-1/num_pareto_points */
  if (used[i] || rand()<(RAND_MAX/(int)num_pareto_points)) {
    /* otherwise use equiprobability */
    size_t pareto_point_index=((unsigned long long)rand()
                                *(unsigned long long)num_pareto_points)
                              /((unsigned long long)(RAND_MAX)+1ull);
    for (i=0; i<numcurrpts; i++) {
      if (pareto_optimal[i]) {
        if (!pareto_point_index--) break;
      }
    }
  }
  best_point=currpts[i];
  used[i]=1;
  printf("Starting local search at x=[");
  if (DIMX) {
    printf("%lf",best_point[0]);
  }
  for (i=1; i<(size_t)DIMX; i++) {
    printf(",%lf",best_point[i]);
  }
  printf("]\n");
  return;
}

void extrapolate_point(void)
{
  DOUBLE cTx[numcurrpts], penalty[numcurrpts];
  size_t i,j;
  for (j=0; j<numcurrpts; j++) {
    cTx[j]=get_point_cTx(currpts[j]);
    penalty[j]=get_point_constraint_violation(currpts[j]);
  }
  /* determine the Pareto-optimal points */
  unsigned char pareto_optimal[numcurrpts];
  size_t num_pareto_points=0;
  memset(pareto_optimal,1,numcurrpts);
  for (i=0; i<numcurrpts; i++) {
    for (j=0; j<numcurrpts; j++) {
      /* if j is strictly better than i, i is not Pareto-optimal */
      if (j!=i && pareto_optimal[j] /* avoid unnecessary comparisons */ &&
          cTx[j] <= cTx[i] && penalty[j] <= penalty[i] &&
          (cTx[j] < cTx[i] || penalty[j] < penalty[i])) {
        pareto_optimal[i]=0;
        goto dont_count;
      }
    }
    num_pareto_points++;
    dont_count:;
  }
  /* throw away the points with excess constraint violation
     also throw away the points which are already feasible up to the tolerance,
     it's no use getting any better there */
  for (i=0; i<numcurrpts; i++) {
    if (pareto_optimal[i]
#ifdef OPTIMUM_TOL
        && (penalty[i]>OPTIMUM_TOL*256. || penalty[i]<OPTIMUM_TOL)
#else
        && (penalty[i]>sqrt(DBL_EPSILON)*256. || penalty[i]<sqrt(DBL_EPSILON))
#endif
       ) {
      pareto_optimal[i]=0;
      num_pareto_points--;
    }
  }
  /* do a ratio reject to get rid of outliers */
  if (num_pareto_points>=3) {
    DOUBLE rho[numcurrpts],rho_mean=0.,rho_var=0.;
    for (i=0; i<numcurrpts; i++) {
      if (pareto_optimal[i]) {
        /* compute rho[i] */
        DOUBLE d_i_NNi=INFINITY, d_NNi_NNNNi=INFINITY;
        size_t NNi;
        /* find nearest neighbor of i and distance */
        for (j=0; j<numcurrpts; j++) {
          if (j!=i && pareto_optimal[j]) {
            DOUBLE d_i_j=0;
            size_t k;
            for (k=0; k<(size_t)DIMX; k++) {
              d_i_j+=(currpts[j][k]-currpts[i][k])*(currpts[j][k]-currpts[i][k]);
            }
            if (d_i_j<d_i_NNi) {
              d_i_NNi=d_i_j;
              NNi=j;
            }
          }
        }
        /* find nearest neighbor of NNi and distance */
        for (j=0; j<numcurrpts; j++) {
          if (j!=i && j!=NNi && pareto_optimal[j]) {
            DOUBLE d_NNi_j=0;
            size_t k;
            for (k=0; k<(size_t)DIMX; k++) {
              d_NNi_j+=(currpts[j][k]-currpts[NNi][k])*(currpts[j][k]-currpts[NNi][k]);
            }
            if (d_NNi_j<d_NNi_NNNNi) {
              d_NNi_NNNNi=d_NNi_j;
            }
          }
        }
        /* rho[i] is the ratio */
        rho[i]=d_i_NNi/d_NNi_NNNNi;
        rho_mean+=rho[i];
      }
    }
    /* compute mean and variance */
    rho_mean/=num_pareto_points;
    for (i=0; i<numcurrpts; i++) {
      if (pareto_optimal[i]) {
        rho_var+=(rho[i]-rho_mean)*(rho[i]-rho_mean);
      }
    }
    rho_var/=num_pareto_points;
    /* ratio-reject cutoff = mean + 3*stddev */
    DOUBLE cutoff=rho_mean+3*sqrt(rho_var);
    /* now reject the outliers */
    for (i=0; i<numcurrpts; i++) {
      if (pareto_optimal[i] && rho[i]>cutoff) {
        printf("Extrapolation: outlier cv=%lf, x=[",penalty[i]);
        if (DIMX) {
          printf("%lf",currpts[i][0]);
        }
        for (j=1; j<(size_t)DIMX; j++) {
          printf(",%lf",currpts[i][j]);
        }
        printf("] rejected (rho=%lf>%lf)\n",rho[i],cutoff);
        pareto_optimal[i]=0;
        num_pareto_points--;
      }
    }
  }
  /* compute the cubic regression and extrapolate to 0 using Cramer's rule */
  DOUBLE S0=0.,S1=0.,S2=0.,S3=0.,S4=0.,S5=0.,S6=0.;
  DOUBLE Sx0[DIMX]={},Sx1[DIMX]={},Sx2[DIMX]={},Sx3[DIMX]={};
  for (i=0; i<numcurrpts; i++) {
    if (pareto_optimal[i]) {
      DOUBLE cv=penalty[i];
      DOUBLE cvn=cv;
      DOUBLE *x=currpts[i];
      DOUBLE cvxn[DIMX];
      printf("Extrapolation considering cv=%lf, x=[",cv);
      if (DIMX) {
        printf("%lf",x[0]);
      }
      for (j=1; j<(size_t)DIMX; j++) {
        printf(",%lf",x[j]);
      }
      printf("]\n");
      S0+=1.;
      for (j=0; j<DIMX; j++) Sx0[j]+=x[j];
      S1+=cv;
      for (j=0; j<DIMX; j++) {
        cvxn[j]=cv*x[j];
        Sx1[j]+=cvxn[j];
      }
      cvn*=cv;
      S2+=cvn;
      for (j=0; j<DIMX; j++) {
        cvxn[j]*=cv;
        Sx2[j]+=cvxn[j];
      }
      cvn*=cv;
      S3+=cvn;
      for (j=0; j<DIMX; j++) {
        cvxn[j]*=cv;
        Sx3[j]+=cvxn[j];
      }
      cvn*=cv;
      S4+=cvn;
      cvn*=cv;
      S5+=cvn;
      cvn*=cv;
      S6+=cvn;
    }
  }
  DOUBLE d1=S6*S4*S2+S5*S3*S4+S4*S5*S3-S4*S4*S4-S3*S3*S6-S2*S5*S5;
  DOUBLE d2=S6*S4*S1+S5*S3*S3+S4*S5*S2-S3*S4*S4-S2*S3*S6-S1*S5*S5;
  DOUBLE d3=S6*S3*S1+S5*S2*S3+S4*S4*S2-S3*S3*S4-S2*S2*S6-S1*S4*S5;
  DOUBLE d4=S5*S3*S1+S4*S2*S3+S3*S4*S2-S3*S3*S3-S2*S2*S5-S1*S4*S4;
  DOUBLE D=S0*d1-S1*d2+S2*d3-S3*d4;
  if (D<DBL_EPSILON*128. && D>-DBL_EPSILON*128.) {
    printf("Extrapolation matrix too ill-conditioned (D=%lg)\n",D);
    return;
  }
  for (j=0; j<DIMX; j++) {
    optimum_x[j]=(Sx0[j]*d1-Sx1[j]*d2+Sx2[j]*d3-Sx3[j]*d4)/D;
  }
  printf("Extrapolation found x=[");
  if (DIMX) {
    printf("%lf",optimum_x[0]);
  }
  for (i=1; i<(size_t)DIMX; i++) {
    printf(",%lf",optimum_x[i]);
  }
  printf("]\n");
  new_point(optimum_x);
}
